/******************************************************************************
 *
 * powerctl_teensy.ino
 *
 * This is the program that runs the telescope power control box. It switches
 * power on and/or off for 8 power ports. Current and voltage monitoring, too.
 *
 *****************************************************************************/

#include <Wire.h>
#include "SparkFun_External_EEPROM.h"

#define LED 13
#define LED_ON 0
#define LED_OFF 1

// Program options 2.
#define OPT_DEW_ENABLE_1 0
#define OPT_DEW_ENABLE_2 1
#define OPT_AUTO_DEW_1 2
#define OPT_AUTO_DEW_2 3

// Command message offsets
#define CMD_OFS 0
#define CMD_TYPE 3
#define CMD_PORT 4
#define CMD_VALUE 6

// Command value types.
#define DT_BOOL 1
#define DT_BYTE 2
#define DT_FLOAT 3

// command and response buffer sizes are the same.
#define BUFFER_SIZE 96

// DMP! command timer
#define DMP_TIME_DELAY 5000

// EEPROM stuff 512 16-byte pages = 8kB
#define EEPROM_SIZE 8192
#define QUEUE_ENTRIES 512
#define EPR_PAGE_SIZE 16
#define EPR_TIME_DELAY 15000

ExternalEEPROM eeprom;

// CUR? Amperage calc period
#define AMP_TIME_DELAY 5000

// Amperage calibration factors
#define CALIBRATION_MULT 1.114
#define CALIBRATION_ADD -0.23

// 11.403A FS, 18.2V FS
// 1.394V/1000 = 0.001394A
// 0.001394A * 8200 = 11.4308 A FS
// 11.4308/4096 = 0.002790723 A per ADU

#define SENSE_FACTOR 8200
#define V_REF 1.402
#define V_PER_ADU V_REF/4096
#define SENSE_A_PER_ADU V_PER_ADU / 1000
#define A_PER_ADU SENSE_A_PER_ADU * SENSE_FACTOR

float ADU2AMPS = A_PER_ADU;


#define VOLTS_PER_ADU 15.422/4096 * 1.098109811
float ADU2VOLTS = VOLTS_PER_ADU;

unsigned long eeprom_tick;
unsigned long amperage_tick;
unsigned long clock_tick;
byte rt_hours = 0;
byte rt_minutes = 0;
byte rt_seconds = 0;

byte state;
bool epr_data_dirty = false;

struct Epr_Data {
    uint16_t seq_no;
    int16_t sig;
    byte options;
    byte power_delays[8];
    byte spare[3];
};

Epr_Data SETUP_DATA{
        .seq_no = 0,
        .sig = 0x0dca,
        .options = 0,
        .power_delays = {0, 0, 0, 0, 0, 0, 0, 0},
        .spare = {0, 0, 0},
};

Epr_Data EPR_DATA {
        .seq_no = 0,
        .sig = 0,
        .options = 0,
        .power_delays = {0, 0, 0, 0, 0, 0, 0, 0},
        .spare = {0, 0, 0},
};

// For clearing the EEPROM.
char ff_16[] = {
        255,255,255,255,255,255,255,255,
        255,255,255,255,255,255,255,255
};

uint32_t last_sequence_no = 0;
uint16_t queue_tail = 0;

int power_pins[] = {0, 1, 2, 3, 4, 5, 6, 7};
int current_pwr_pins[] = {A0, A1, A4, A5, A6, A7, A8, A9};
int voltage_pin = A10;

float current_pwr[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
float ah_pwr[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
float voltage = 0.0;
float calibration_mult_pwr[] = {1.026373, 1.046039, 1.040955, 1.050511, 1.043622, 1.033814, 1.029155, 1.041406};
float calibration_add_pwr[] = {-0.044402, -0.044482, -0.044242, -0.042280, -0.038556, -0.037635, -0.036714, -0.047325};

char port;
bool value_bool;
uint8_t value_byte;
float value_float;

char cmd_buffer[BUFFER_SIZE];
char rsp_buffer[BUFFER_SIZE];
char good_response[] = {"OK\n"};
char version_string[] = {"OK:4.0.1\n"};



/******************************************************************************
 * 
 * load_eeprom  Find and load the latest copy of settings in EEPROM.
 * 
 *****************************************************************************/

void load_eeprom()
{
    int i;
    Epr_Data ed;
    last_sequence_no = 0;
    queue_tail = 0;

    memcpy(&EPR_DATA, &SETUP_DATA, sizeof(EPR_DATA));

    for (i = 0; i < QUEUE_ENTRIES; i++)
    {
        eeprom.get((i * EPR_PAGE_SIZE), ed);
        if ((ed.seq_no > last_sequence_no) && (ed.seq_no != 0xFFFF))
        {
            queue_tail = i;
            last_sequence_no = ed.seq_no;
        }
    }
    eeprom.get((queue_tail * EPR_PAGE_SIZE), EPR_DATA);
}

/******************************************************************************
 * 
 * store_eeprom Store the latest copy in EEPROM and increment counter.
 *
 * Writes to a different page each time it writes, incrementing a 32-bit
 * counter.
 *
 *****************************************************************************/

void store_eeprom() {
    queue_tail++;
    if (queue_tail >= QUEUE_ENTRIES)
        queue_tail = 0;
    last_sequence_no++;
    EPR_DATA.seq_no = last_sequence_no;
    eeprom.put(queue_tail * EPR_PAGE_SIZE, EPR_DATA);
}

/******************************************************************************
 * 
 * erase_eeprom Store ff_32 in block 0, increment counter and repeat to 
 * end of EEPROM.
 *
 *****************************************************************************/

char * erase_eeprom() {
    memcpy(&EPR_DATA, &SETUP_DATA, sizeof(EPR_DATA));
    
    for (int x = 0; x < QUEUE_ENTRIES; ++x) {
        eeprom.put(x * EPR_PAGE_SIZE, ff_16);
    }
    eeprom.put(0, EPR_DATA);
    last_sequence_no = 0;
    strcpy(rsp_buffer, "OK\n");
    return rsp_buffer;
}

/******************************************************************************
 *
 * reset_totals Clears all of the accumulated stats
 *
 *****************************************************************************/

char * reset_totals(void) {
    for (int x = 0; x < 8; ++x) {
        ah_pwr[x] = 0.0;
    }
    rt_hours = 0;
    rt_minutes = 0;
    rt_seconds = 0;
    return good_response;
}

/******************************************************************************
 *
 * setPortPwr   Turns the power on for this port.
 *
 * @param p     The port number (0-7).
 * @param flag  The port should be powered on - True or false.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * setPortPwr(byte p, bool flag) {
    if (flag) {
        state = bitSet(state, p);
    } else {
        state = bitClear(state, p);
    }
    pinMode(power_pins[p], OUTPUT);
    digitalWrite(power_pins[p], flag ? HIGH : LOW);
    return good_response;
}

/******************************************************************************
 *
 * getPortPwr   Returns the status of the port pin.
 *
 * @param p     The port to query.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * getPortPwr(byte p) {
    strcpy(rsp_buffer, "OK:");
    strcat(rsp_buffer, bitRead(state, p) ? "ON\n" : "OFF\n");
    return rsp_buffer;
}

/******************************************************************************
 *
 * setPortAuto  Sets the port to automatically come on at power up.
 *
 * @param p     The port to set.
 * @param flag  The value to set - True or false.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * setPortAuto(byte p, bool flag) {
    if (flag) {
        EPR_DATA.options = bitSet(EPR_DATA.options, p);
    } else {
        EPR_DATA.options = bitClear(EPR_DATA.options, p);
    }
    epr_data_dirty = true;
    return good_response;
}

/******************************************************************************
 *
 * getPortAuto  Gets the auto start flag for the specified port.
 *
 * @param p     The port to query.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * getPortAuto(byte p) {
    strcpy(rsp_buffer, "OK:");
    strcat(rsp_buffer, bitRead(EPR_DATA.options, p) ? "ON\n" : "OFF\n");
    return rsp_buffer;
}

/******************************************************************************
 *
 * setPowerDelay        Set the delay before powering up this pin. Delay does
 *                      not apply to low outputs - only high outputs.
 *
 * @param p     Port number (0 - 5)
 * @param v     Value (0 to 255 == (0mS - 512mS)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * setPowerDelay(byte p, byte v) {
    if (p > 7) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    if (EPR_DATA.power_delays[p] != v) {
        EPR_DATA.power_delays[p] = v;
        epr_data_dirty = true;
    }
    return good_response;
}

/******************************************************************************
 *
 * getPowerDelay        Get the delay before powering up this pin.
 *
 * @param p     Port number (0 - 5)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * getPowerDelay(byte p) {
    if (p > 7) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03d\n", EPR_DATA.power_delays[p]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * @param p switched port number
 * @return float amperage
 *
 *****************************************************************************/

float _get_current_pwr(byte p) {
    float amps;

    amps = analogRead(current_pwr_pins[p]);
    delay(1);
    amps += analogRead(current_pwr_pins[p]);
    delay(1);
    amps += analogRead(current_pwr_pins[p]);
    delay(1);
    amps += analogRead(current_pwr_pins[p]);
    amps *= 0.25;
    amps = (amps * ADU2AMPS + calibration_add_pwr[p]) * calibration_mult_pwr[p];
    if (amps < 0.005) amps = 0.0;
    current_pwr[p] = amps;
    ah_pwr[p] += amps / 720.0;
    return amps;
}

/******************************************************************************
 *
 * _get_voltage()       Get the applied voltage (12V or so)
 *
 * @return float voltage
 *
 *****************************************************************************/

 float _get_voltage() {
    float volts;

    volts = analogRead(voltage_pin);
    delay(1);
    volts += analogRead(voltage_pin);
    delay(1);
    volts += analogRead(voltage_pin);
    delay(1);
    volts += analogRead(voltage_pin);
    voltage = volts / 4 * ADU2VOLTS;  // 16V/4096.
    return voltage;
}

/******************************************************************************
 *
 * getPwrCurrent    Get the current used by a port.
 *
 * @param p     Port number (0 - 7)
 * @return      good response or error message.
 *
 * 3.3V = 4095
 * 3.3V = 33Amps
 * 1 ADU = 33/4096 Amps = 0.008056641 A per ADU
 * 3 samples: (ADU x 3) / 0.024169922 = Amps.
 *
 *****************************************************************************/

char * getPwrCurrent(byte p) {
    float amps;

    if (p > 7) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    amps = current_pwr[p];
    if (amps < 0.0) amps = 0.0;
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.3f\n", amps);
    return rsp_buffer;
}

/******************************************************************************
 *
 * get 
 * rentTotal  API call to get total current
 *
 * @return          float amperage
 *
 *****************************************************************************/

char * getCurrentTotal() {
    float amps = 0.0;

    for (int x = 0; x < 8; ++x) {
        if (bitRead(state, x))
            amps += current_pwr[x];
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.3f\n", amps);
    return rsp_buffer;
}

char * getAHTotal() {
    float ah = 0.0;

    for (int x = 0; x < 8; ++x) {
        ah += ah_pwr[x];
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", ah);
    return rsp_buffer;
}

char * getPortAH(byte p) {
    if (p > 7) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", ah_pwr[p]);
    return rsp_buffer;
}

char * getRunTime() {

    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03u:%02d:%02d\n",
             rt_hours, rt_minutes, rt_seconds);
    return rsp_buffer;
}

char * getADCValues(byte p) {
    float amps;
    int adu;
    adu = analogRead(current_pwr_pins[p]);
    delay(2);
    adu += analogRead(current_pwr_pins[p]);
    delay(2);
    adu += analogRead(current_pwr_pins[p]);
    adu /= 3;
    amps = adu * ADU2AMPS;

    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%04d  %4.3f %8.7f\n", adu, amps, ADU2AMPS);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getVoltageInput  API call to get input voltage (9V-14V)
 *
 * @return          float voltage
 *
 *****************************************************************************/

char * getVoltageInput() {
    float volts;
    
    volts = _get_voltage();
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", volts);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getPort      Gets the port number from the host message.
 *
 * @return      True on success, false otherwise.
 *
 *****************************************************************************/

bool getPort() {
    port = cmd_buffer[CMD_PORT] - '0';
    if (port < 0 || port > 7)
        return false;
    return true;
}

/******************************************************************************
 *
 * getPortAndValue  Get the channel number and value from the message.
 *
 * @param d         The type of value to expect. bool, byte, byte.
 * @return          True on success, false otherwise.
 *
 *****************************************************************************/

bool getPortAndValue(int d) {
    // Port is one digit, and value is a bool, byte or float.
    port = cmd_buffer[CMD_PORT] - '0';
    if (port < 0 || port > 7)
        return false;
    int tmp;
    switch (d) {
        case DT_BOOL:
            value_bool = cmd_buffer[CMD_VALUE] & 1;
            break;
        case DT_BYTE:
            tmp = atoi((const char *) &cmd_buffer[CMD_VALUE]);
            if (tmp > 255) {tmp = 255;}
            else if (tmp < 0) {tmp = 0;}
            value_byte = (byte)tmp;
            break;
        case DT_FLOAT:
            value_float = atof((const char *) &cmd_buffer[CMD_VALUE]);
            break;
        default:
            return false;
    }
    return true;
}

/******************************************************************************
 *
 * parseCommand     Takes the command apart and calls the appropriate action.
 *
 * @return          None.
 *
 *****************************************************************************/

void parseCommand() {
    if (!strncmp(cmd_buffer, "PWR", 3)) {
        // Command PWR sets or gets the state to ON (true) or OFF (false).
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BOOL)) {
                Serial.print(setPortPwr(port, value_bool));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPortPwr(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "PWA", 3)) {
        // Command PWA sets or gets the auto on at power up flag.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BOOL)) {
                Serial.print(setPortAuto(port, value_bool));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPortAuto(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "DLY", 3)) {
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BYTE)) {
                Serial.print(setPowerDelay(port, value_byte));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPowerDelay(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "PAH", 3)) {
        // Command PAH sets or gets the switched port amp hours accumulator.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_FLOAT)) {
                //Serial.print(setPortAH(port, value_float));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPortAH(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }

    } else if (!strncmp(cmd_buffer, "ADC", 3)) {
        // Get the current temperature.
        if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getADCValues(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "VER", 3)) {
        // Get the program version number.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(version_string);
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "CUR", 3)) {
        // Get the total current.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getCurrentTotal());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "AHT", 3)) {
        // Get the AmpHr total.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getAHTotal());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "VIN", 3)) {
        // Get the input voltage.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getVoltageInput());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "PWC", 3)) {
        // Get the current used by port.
        if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPwrCurrent(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "ERA", 3)) {
        // Erase the EEPROM.
        if (cmd_buffer[CMD_TYPE] == '#') {
            Serial.print(erase_eeprom());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "CLR", 3)) {
        // Reset run timer and all totals.
        if (cmd_buffer[CMD_TYPE] == '#') {
            Serial.print(reset_totals());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "EPR", 3)) {
        // Dump the current EEPROM config.
        if (cmd_buffer[CMD_TYPE] == '?') {
            print_eeprom(&EPR_DATA);
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "RNT", 3)) {
        // Dump the current run timer.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getRunTime());
        } else {
            Serial.print("ER:cmd type\n");
        }
    }
}

/******************************************************************************
 *
 * print_eeprom      Prints the EEPROM block to the Serial port.
 *
 *****************************************************************************/

void print_eeprom(Epr_Data * ptr) {
    byte x;
    Serial.print("SizeOf :  ");
    Serial.println(sizeof(EPR_DATA));
    Serial.print("SeqNo  :  ");
    Serial.println(ptr->seq_no);
    Serial.print("Sig    :  ");
    Serial.println(ptr->sig, 16);
    Serial.print("Opt    :  ");
    Serial.println(ptr->options, 2);
    Serial.print("Delays :  ");
    for (x=0;x<8;++x) {
        Serial.print(ptr->power_delays[x]);
        Serial.print("  ");
    }
    Serial.println("");
}

/******************************************************************************
 *
 * setup        Setup the hardware and firmware to run.
 *
 *****************************************************************************/

void setup() {

    Serial.begin(230400);
    pinMode(LED, OUTPUT);

    Wire.begin();
    Wire.setClock(400000);
    Wire.setSDA(17);
    Wire.setSCL(16);

    interrupts();
    load_eeprom();

    // Set the starting value of all ports, with delays.
    for (byte x = 0; x < 8; ++x) {
        pinMode(power_pins[x], OUTPUT);
        digitalWrite(power_pins[x], LOW);
        if (bitRead(EPR_DATA.options, x)) {
            delay(EPR_DATA.power_delays[x] * 4);
            digitalWrite(power_pins[x], HIGH);
            state |= bit(x);
        } else {
            state &= ~bit(x);
        }
    }
    analogReference(EXTERNAL);
    analogReadRes(12);
    memset(cmd_buffer, 0, sizeof(cmd_buffer));
    eeprom_tick = millis() + EPR_TIME_DELAY;
    amperage_tick = millis() + AMP_TIME_DELAY;
    clock_tick = millis() + 1000;
    rt_hours = 0;
    rt_minutes = 0;
    rt_seconds = 0;
}

/******************************************************************************
 * loop     Main program loop.
 *
 * Checks for a command every mS, checks if EEPROM needs writing, current 
 * needs reading, or the RTC needs updating.
 *
 *****************************************************************************/

void loop() {

    // Did the host request something?
    if (Serial.available()) {
        // Read until linefeed.
        if (Serial.readBytesUntil(10, cmd_buffer, sizeof(cmd_buffer))) {
            parseCommand();
            memset(cmd_buffer, 0, sizeof(cmd_buffer));
        }
    }
    // Is it time to read current?
    if (amperage_tick < millis()) {
        amperage_tick = millis() + AMP_TIME_DELAY;
        for (byte idx = 0; idx < 8; ++idx) {
            current_pwr[idx] = _get_current_pwr(idx);
        }
    }
    // Is it time to flush to EEPROM?
    if (eeprom_tick < millis()) {
        eeprom_tick = millis() + EPR_TIME_DELAY;
        if (epr_data_dirty) {
            store_eeprom();
            epr_data_dirty = false;
        }
    }
    // Is it time to update the RTC?
    if (clock_tick < millis()) {
        clock_tick = millis() + 1000;
        if(++rt_seconds > 59) {
            rt_seconds = 0;
            if (++rt_minutes > 59) {
                rt_minutes = 0;
                ++rt_hours;
            }
        }
    }
}
